#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <utime.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

#include "sftp.h"

#define skipspaces(s) {while (s && *s && isspace(*s)) s++;}

extern volatile int cancelled;
extern u_int32_t version;
extern int verbose;

int
next_token(char *str, char **arg) {
	char *s, *orig, *startq;
	int escaped = 0, quoted = 0;

	orig = s = str;
	startq = NULL;
	while (1) {
		if (*s == 0) {
			*arg = str;
			return (s - orig);
		}
		if (escaped) {
			memmove(str + 1, str, s - str - 1);
			str++;
			escaped = 0;
		}
		else if (quoted) {
			if (*s == '"') {
				memmove(startq + 2, startq + 1, s - startq - 1);
				memmove(str + 2, str, startq - str);
				str += 2;
				quoted = 0;
			}
		}
		else {
			if (*s == '"') {
				quoted = 1;
				startq = s;
			}
			else if (*s == '\\')
				escaped = 1;
			else if (isspace(*s))
				break;
		}
		s++;
	}
	*s++ = 0;
	*arg = str;
	return (s - orig);
}

int
tokenize(char *str, char **args, int maxargs) {
	int i;

	for (i = 0; str && *str && i < maxargs - 1; i++) {
		skipspaces(str);
		if (*str == 0)
			break;
		str += next_token(str, &args[i]);
	}
	args[i] = NULL;
	return i;
}

int
find_match(char *str, map *action) {
	int i, guess = -1, badguess = 0;
	for (i=0; action[i].str; i++) {
		if (strcmp(action[i].str, str) == 0)
			return i;
		if (strncmp(action[i].str, str, strlen(str)) == 0) {
			if (guess < 0 && !badguess)
				guess = i;
			else {
				badguess = 1;
				guess = -1;
			}
		}
	}
	return guess;
}

int
do_action(char *str, map *action) {
	int index, nargs;
	char *args[MAXARGS];

	if (str[strlen(str)-1] == '\n')
		str[strlen(str)-1] = 0;
	skipspaces(str);
	if (*str == 0)
		return 0;
	if (*str == '!') {
		args[0] = "!";
		args[1] = str + 1;
		nargs = 2;
        }
	else 
		nargs = tokenize(str, args, MAXARGS);
	if (nargs == 0)
		return 0;
	index = find_match(args[0], action);
	if (index >= 0)
		return action[index].func(args, nargs);
	else
		printf ("%s: invalid command\n", args[0]);
	return -1;
}

int
do_sendfile(int sock, char *file, sftp_transfer_callback callback) {
	struct stat statbuf;
	u_int32_t size, mtime;
	u_int16_t mode;
	char modestr[9];
	int n, fd;
	char buf[BUFSIZE];
	char *base;
	message m;
	char *str;
	u_int8_t c = 1;
	u_int32_t skipped = 0;
	sftp_transfer_info info;

	memset(&info, 0, sizeof(info));
	info.name = file;
	callback(&info);

	if (stat(file, &statbuf) || !S_ISREG(statbuf.st_mode)) {
		str = "File doesn't exist or is not a regular file";
		send_message(sock, _data_message(c, ERROR, str, strlen(str)+1));
		return -1;
	}
	fd = open(file, O_RDONLY);
	if (fd < 0) {
		str = "File could not be opened";
		send_message(sock, _data_message(c, ERROR, str, strlen(str)+1));
		return -1;
	}

	base = findbasename(file);
	mode = statbuf.st_mode & (S_IRWXU|S_IRWXG|S_IRWXO);
	modetostr(mode, modestr);
	info.size = statbuf.st_size;
	size = htonl(statbuf.st_size);
	mtime = htonl(statbuf.st_mtime);

	send_message(sock, _message(SENDFILE, &c, 1));
	send_message(sock, _data_message(c, FILENAME, base, strlen(base)+1));
	n = recv_message(sock, &m);

	if (n < 0)
		return -1;

	if (!(m.command == FILEOK && m.len == 1 && *(char *)m.data != 0)) {
		xfree(m.data);
		return -1;
	}

	xfree(m.data);
	send_message(sock, _data_message(c, FILESIZE, &size, 4));
	send_message(sock, _data_message(c, FILEMODE, modestr, 9));
	if (version >= SFTP_VERSION(0,7,0))
		send_message(sock, _data_message(c, FILETIME, &mtime, 4));

	if (skipped > 0) {
		lseek(fd, skipped, SEEK_SET);
		info.start = skipped;
		skipped = htonl(skipped);
		send_message(sock, _data_message(c, SKIPBYTES, &skipped, 4));
	}
	while ((n = read(fd, buf, BUFSIZE)) > 0) {
		if (cancelled)
			break;
		send_message(sock, _data_message(c, DATA, buf, n));
		info.increment = n;
		info.total += n;
		callback(&info);
	}
	send_message(sock, _data_message(c, ENDDATA, NULL, 0));
	info.increment = 0;
	callback(&info);
	close (fd);
	return 0;
}

int
do_recvfile(int sock, u_int8_t channel, int resume,
	    sftp_transfer_callback callback)
{
	int n;
	unsigned int flags;
	u_int32_t mtime = 0;
	u_int16_t mode;
	int fd = -1;
	u_int8_t status;
	message m;
	int sentcancel = 0;
	sftp_transfer_info info;

	memset(&info, 0, sizeof(info));
	while (1) {
		n = recv_message(sock, &m);
		if (n < 0)
			return -1;
		switch (m.command) {
			case SENDFILE:
				channel = *(u_int8_t*)m.data;
			case FILENAME:
				info.name = strdup(m.data);
				flags = O_WRONLY|O_CREAT;
				if (!resume)
					flags |= O_TRUNC;
				fd = open(info.name, flags, 0644);
				callback(&info);
				status = (fd >= 0);
				if (status > 0 && resume) {
					struct stat statbuf;
					u_int32_t nsize;
					fstat(fd, &statbuf);
					info.start = statbuf.st_size;
					info.total = info.start;
					lseek(fd, info.start, SEEK_SET);
					nsize = htonl(info.start);
					send_message(sock,
						     _data_message(channel,
								   SKIPBYTES,
								   &nsize, 4));
				}
				else {
					send_message(sock,
						     _data_message(channel,
								   FILEOK,
								   &status, 1));
				}
				if (status == 0) {
					free(info.name);
					return -1;
				}
				break;
			case FILESIZE:
				if (fd < 0 || m.len != 4)
					break;
				info.size = ntohl(*(u_int32_t *)m.data);
				if (verbose)
					printf("%d bytes\n%s: ",
					       info.size, info.name);
				break;
			case FILEMODE:
				if (fd < 0 || m.len != 9)
					break;
				mode = strtomode((char *)m.data);
				fchmod(fd, mode);
				break;
			case FILETIME:
				if (fd < 0 || m.len != 4)
					break;
				mtime = ntohl(*(u_int32_t *)m.data);
				break;
			case DATA:
				if (fd < 0)
					break;
				write(fd, m.data, m.len);
				info.increment = m.len;
				info.total += m.len;
				callback(&info);
				break;
			case ENDDATA:
				close(fd);
				if (mtime != 0) {
					struct utimbuf ubuf;
					ubuf.actime = time(NULL);
					ubuf.modtime = mtime;
					utime(info.name, &ubuf);
				}
				info.increment = 0;
				callback(&info);
				free(info.name);
				return info.total - info.start;
		}
		xfree(m.data);
		if (cancelled && !sentcancel &&
		    version >= SFTP_VERSION(0, 8, 0))
		{
			send_message(sock,
				     _data_message(channel, CANCEL, NULL, 0));
			sentcancel = 1;
		}
	}
}
